"""Python CACAO Stix Class Definitions """

from collections import OrderedDict

from stix2.exceptions import (
    PropertyPresenceError, )
from stix2.properties import (
    BooleanProperty, ExtensionsProperty, IDProperty, IntegerProperty, ListProperty, OpenVocabProperty, ReferenceProperty, StringProperty,
    TimestampProperty, TypeProperty,
)
from stix2.utils import NOW
from stix2.v21.base import _DomainObject, _STIXBase21
from stix2.v21.common import (
    ExternalReference, GranularMarking, KillChainPhase,
)
from stix2.v21.vocab import (
    ATTACK_MOTIVATION, ATTACK_RESOURCE_LEVEL, IMPLEMENTATION_LANGUAGE, MALWARE_CAPABILITIES, MALWARE_TYPE,
    PROCESSOR_ARCHITECTURE, TOOL_TYPE,
)

import logging
logger = logging.getLogger(__name__)




import json
import os
from enum import Enum
from typing import Optional

from pydantic import BaseModel

class DefinitionName(str, Enum):
    ATTACK = "attack"
    CACAO = "cacao"
    KESTREL = "kestrel"
    OS_THREAT = "os_threat"
    STIX_21 = "stix21"
    US_DoD = "us_dod"


class ImportTypeToDefinitionMapper():

    @staticmethod
    def corresponding_definition_name(import_type: str) -> Optional[DefinitionName]:
        lookup = {
            "ATTACK": DefinitionName.ATTACK,
            "CACAO": DefinitionName.CACAO,
            "kestrel": DefinitionName.KESTREL,
            "os_threat": DefinitionName.OS_THREAT,
            "STIX21": DefinitionName.STIX_21,
            "US_DoD": DefinitionName.US_DoD,
        }
        return lookup.get(import_type, None)

class ObjectKeys(str, Enum):
    SDO = "sdo"
    SRO = "sro"
    SCO = "sco"
    SUB = "sub"
    META = "meta"

class ModelClassDefinition(BaseModel):
    sdo: dict
    sro: dict
    meta: dict
    sub: dict
    sco: dict







class ModelDefinition(BaseModel):
    base: dict
    mappings: dict
    sub_objects: dict
    data: dict
    class_def: ModelClassDefinition


import copy
from datetime import timezone

from stixorm.module.authorise import authorised_mappings

import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType

logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------------------------
#  1. Convert TypeQl Ans to Res
# --------------------------------------------------------------------------------------------------------


def convert_ans_to_res(answer_iterator, r_tx, import_type: ImportType):
    """
    Take the response from TypeDB to a query, and start the process to use GRPC (typedb_lib-client) commands
    to expand on the returned data to some potential object shape (i.e. mandatory and optional),
    and report it back in some intermediate form that can then be reprocessed into Stix 2.1 format.
    This depends on the shape of the Stix object involved.

    The first step is to branch depending on whether we are decoding an entity or relation.

    Args:
        answer_iterator (): current answers
        r_tx (): transaction thread to query
        import_type (): stix2.1 or att&ck, used in the second half

    Returns:
        res: A list of data objects, in the intermediate form for processing into Stix objects
    """
    res = []

    for answer in answer_iterator:
        dict_answer = answer.map()
        for key, thing in dict_answer.items():
            # pull entity data
            if thing.is_entity():
                # 1. describe entity
                ent = {'type': 'entity', 'symbol': key, 'T_id': thing.get_iid(),
                       'T_name': thing.get_type().get_label().name()}
                # 2 get and dsecribe properties
                props_obj = thing.as_remote(r_tx).get_has()
                ent['has'] = process_props(props_obj)
                # 3. get and describe relations
                reln_types = thing.as_remote(r_tx).get_relations()
                ent['relns'] = process_relns(reln_types, r_tx, import_type)
                res.append(ent)
                # logger.debug(f'ent -> {ent}')

            # pull relation data
            elif thing.is_relation():
                # 1. setup basis
                rel = {'type': 'relationship', 'symbol': key, 'T_id': thing.get_iid(),
                       'T_name': thing.get_type().get_label().name()}
                att_obj = thing.as_remote(r_tx).get_has()
                rel['has'] = process_props(att_obj)
                # 3. get and describe relations
                reln_types = thing.as_remote(r_tx).get_relations()
                rel['relns'] = process_relns(reln_types, r_tx, import_type)
                # 4. get and describe the edges
                edges = []
                edge_types = thing.as_remote(r_tx).get_players_by_role_type()
                stix_id = r_tx.concepts().get_attribute_type("stix-id")
                for role, things in edge_types.items():
                    edge = {"role": role.get_label().name(), 'player': []}
                    for thing in things:
                        if thing.is_entity():
                            edge['player'].append(process_entity(thing, r_tx,stix_id))

                    edges.append(edge)

                rel['edges'] = edges
                res.append(rel)

            # else log out error condition
            else:
                logger.debug(f'Error key is {key}, thing is {thing}')

    return res


def process_entity(thing, r_tx, stix_id: str):
    """
        If the current returned object from typedb_lib contains an entity then unpack it using grpc commands
        into an interim list
    Args:
        thing (): the grpc entity reference
        r_tx (): the typedb_lib transaction
        stix_id (): the stix object id

    Returns:
        play {}: a return dict
    """
    play = {"type": "entity", "tql": thing.get_type().get_label().name()}
    attr_stix_id = thing.as_remote(r_tx).get_has(attribute_type=stix_id)
    for attr in attr_stix_id:
        play["stix_id"] = attr.get_value()

    return play


def process_relns(reln_types, r_tx, import_type: ImportType):
    """
        If the current returned object is a list of relations (i.e. a list of embedded objects), then unpack them
    Args:
        reln_types (): iterable of relation types
        r_tx (): returned transaction

    Returns:
        relns []: a list of reln's
    """
    relns = []
    for r in reln_types:
        logger.debug(f'r  - >{r}, rtx -> {r_tx}')
        reln = get_relation_details(r, r_tx, import_type)
        relns.append(reln)

    return relns


def process_relation(p, r_tx, stix_id):
    """
        If the current returned object is a relation (i.e. embedded object) then unpack it
    Args:
        p ():  returned object
        r_tx (): returned transaction
        stix_id (): the id of the stix object

    Returns:
        plays {}: a dict containing the unpacked relation
    """
    plays = {"type": "attribute", "tql": p.get_type().get_label().name()}
    attr_stix_id = p.as_remote(r_tx).get_has(attribute_type=stix_id)
    for attr in attr_stix_id:
        plays["stix_id"] = attr.get_value()

    return plays


def process_props(props_obj):
    """
        Unpack  a list of properties/values
    Args:
        props_obj (): iterable object of properties

    Returns:
        props []: a list of properties
    """
    props = []
    for a in props_obj:
        prop = {"typeql": a.get_type().get_label().name()}
        if a.is_datetime():
            nt_obj = a.get_value()
            dt_obj = nt_obj.astimezone(timezone.utc)
            prop["value"] = dt_obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            prop['datetime'] = True
        elif a.is_string():
            temp_string = a.get_value()
            prop["value"] = temp_string.replace("\\\\", "\\")
            prop['datetime'] = False
        else:
            prop["value"] = a.get_value()
            prop['datetime'] = False

        props.append(prop)

    return props


def process_value(p):
    """
        If object is a value, then unpack it
    Args:
        p (): an object that is a value

    Returns:
        ret_value : a returned value
    """
    if p.is_datetime():
        nt_obj = p.get_value()
        dt_obj = nt_obj.astimezone(timezone.utc)
        ret_value = dt_obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    else:
        ret_value = p.get_value()

    return ret_value


def get_relation_details(r, r_tx, import_type: ImportType):
    """
        For a given sub-object type, unpack it
    Args:
        r (): the embedded relation object
        r_tx (): the retrned transaction

    Returns:
        reln {}: a dict containing the reln details
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    reln = {}
    reln_name = r.get_type().get_label().name()
    reln['T_name'] = reln_name
    reln['T_id'] = r.get_iid()
    if reln_name in auth["tql_types"]["embedded_relations"]:
        reln['roles'] = get_embedded_relations(r, r_tx)

    elif reln_name in auth["tql_types"]["standard_relations"] or reln_name == "sighting":
        reln['roles'] = get_standard_relations(r, r_tx)

    elif reln_name in auth["tql_types"]["key_value_relations"]:
        reln['roles'] = get_key_value_relations(r, r_tx)

    elif reln_name in auth["tql_types"]["extension_relations"]:
        reln['roles'] = get_extension_relations(r, r_tx, import_type)

    elif reln_name in auth["tql_types"]["list_of_objects"]:
        reln['roles'] = get_list_of_objects(r, r_tx, import_type)

    elif reln_name == "granular-marking":
        reln['roles'] = get_granular_marking(r, r_tx)

    elif reln_name == "hashes" or reln_name == "file-header-hashes":
        reln['roles'] = get_hashes(r, r_tx)

    else:
        logger.error(f'Error, relation name is {reln_name}')

    return reln


def reln_map_entity_attribute(reln_map, r_tx, stix_id: str, is_kv):
    """
        Process a map of Player by Role types, and unpack entity and attribute
    Args:
        reln_map (): relation map of player by role types
        r_tx (): transaction
        stix_id (): stix id
        is_kv (): do extra stuff if coming froma key-value relationship

    Returns:
        roles []: list of dict objects
    """
    roles = []
    for role, player in reln_map.items():
        role_i = {'role': role.get_label().name(), 'player': []}
        for p in player:
            play = {}
            if p.is_entity():
                role_i['player'].append(process_entity(p, r_tx, stix_id))
            elif p.is_attribute():
                play["type"] = "attribute"
                play["tql"] = p.get_type().get_label().name()
                play["value"] = process_value(p)
                if is_kv:
                    att_obj = p.as_remote(r_tx).get_has()
                    play['props'] = process_props(att_obj)
                role_i['player'].append(play)

            else:
                logger.debug(f'player is not entity type {p}')

        roles.append(role_i)

    return roles


def get_granular_marking(r, r_tx):
    """
        Process a granular marking sub object through grpc
    Args:
        r (): the typedb_lib object
        r_tx (): the transaction

    Returns:
        roles []: list of dict objects
    """
    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()
    is_kv: object = False
    roles = reln_map_entity_attribute(reln_map, r_tx, stix_id, is_kv)
    return roles


def get_hashes(r, r_tx):
    """
        Process a get hashes sub object through grpc
    Args:
        r (): the typedb_lib object
        r_tx (): the transaction

    Returns:
        roles []: list of dict objects
    """
    roles = []
    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    hash_value = r_tx.concepts().get_attribute_type("hash-value")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()

    for role, player in reln_map.items():
        role_name = role.get_label().name()
        role_i = {'role': role_name, 'player': []}
        for p in player:
            play = {}
            if p.is_entity():
                play["type"] = "entity"
                play["tql"] = p.get_type().get_label().name()
                if role_name == "owner":
                    attr_stix_id = p.as_remote(r_tx).get_has(attribute_type=stix_id)
                    for attr in attr_stix_id:
                        play["stix_id"] = attr.get_value()
                else:
                    attr_hash_value = p.as_remote(r_tx).get_has(attribute_type=hash_value)
                    for attr in attr_hash_value:
                        play["hash_value"] = attr.get_value()

                role_i['player'].append(play)

            else:
                logger.debug(f'player is not entity type {p}')

        roles.append(role_i)
    return roles


def get_key_value_relations(r, r_tx):
    """
        Process a key-value sub object through grpc
    Args:
        r (): the typedb_lib object
        r_tx (): the transaction

    Returns:
        roles []: list of dict objects
    """
    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()
    is_kv: object = True
    roles = reln_map_entity_attribute(reln_map, r_tx, stix_id, is_kv)
    return roles


def get_list_of_objects(r,
                        r_tx,
                        import_type: ImportType):
    """
        Process a list of objects sub object through grpc
    Args:
        r (): the typedb_lib object
        r_tx (): the transaction

    Returns:
        roles []: list of dict objects
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    reln_name = r.get_type().get_label().name()
    for lot in auth["reln"]["list_of_objects"]:
        if reln_name == lot["typeql"]:
            reln_pointed_to = lot["pointed_to"]
            reln_object = lot["object"]
            reln_object_props = copy.deepcopy(auth["sub_objects"][reln_object])
            reln_stix = lot["name"]

    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()
    roles = []
    for role, player in reln_map.items():
        role_i = {'role': role.get_label().name(), 'player': []}
        for p in player:
            play = {}
            if p.is_entity():
                play["type"] = "entity"
                play["tql"] = p.get_type().get_label().name()
                props_obj = p.as_remote(r_tx).get_has()
                play['has'] = process_props(props_obj)
                # 3. get and describe relations
                reln_types = p.as_remote(r_tx).get_relations()
                relns = []
                for rel in reln_types:
                    reln = {}
                    reln_name = rel.get_type().get_label().name()

                    reln['T_name'] = reln_name
                    reln['T_id'] = rel.get_iid()
                    reln_map = rel.as_remote(r_tx).get_players_by_role_type()
                    reln['roles'] = reln_map_entity_relation(reln_map, r_tx, stix_id)
                    relns.append(reln)

                play['relns'] = relns
                role_i['player'].append(play)


            else:
                logger.debug(f'player is not entity type {p}')

        roles.append(role_i)
    return roles


def reln_map_entity_relation(reln_map,
                             r_tx,
                             stix_id: str):
    """

    Args:
        reln_map ():
        r_tx ():
        stix_id ():

    Returns:

    """
    roles = []
    for role, player in reln_map.items():
        role_i = {'role': role.get_label().name(), 'player': []}
        for p in player:
            play = {}
            if p.is_entity():
                role_i['player'].append(process_entity(p, r_tx, stix_id))
            elif p.is_relation():
                role_i['player'].append(process_relation(p, r_tx, stix_id))

            else:
                logger.debug(f'player is not entity type {p}')

        roles.append(role_i)

    return roles


def get_embedded_relations(r, r_tx):
    """
        Process embedded relationships (i.e. based on Stix-id)
    Args:
        r (): relation
        r_tx (): transaction

    Returns:
        roles []: list of dict objects
    """
    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()
    roles = reln_map_entity_relation(reln_map, r_tx, stix_id)
    return roles


def get_extension_relations(r,
                            r_tx,
                            import_type: ImportType):
    """
        Process a Stix extension sub object through grpc
    Args:
        r (): the typedb_lib object
        r_tx (): the transaction

    Returns:
        roles []: list of dict objects
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    reln_name = r.get_type().get_label().name()
    for ext in auth["reln"]["extension_relations"]:
        if ext['relation'] == reln_name:
            reln_object = ext['object']

    stix_id = r_tx.concepts().get_attribute_type("stix-id")
    reln_map = r.as_remote(r_tx).get_players_by_role_type()
    roles = []
    for role, player in reln_map.items():
        role_i = {'role': role.get_label().name(), 'player': []}
        for p in player:
            play = {}
            if p.is_entity():
                play["type"] = "entity"
                p_name = p.get_type().get_label().name()
                play["tql"] = p_name
                if p_name == reln_object:
                    props_obj = p.as_remote(r_tx).get_has()
                    play['has'] = process_props(props_obj)
                    # 3. get and describe relations
                    reln_types = p.as_remote(r_tx).get_relations()
                    relns = []
                    for rel in reln_types:
                        reln = {}
                        reln = validate_get_relns(rel, r_tx, reln_object, import_type)
                        if reln == {} or reln is None:
                            pass
                        else:
                            relns.append(reln)

                    play['relns'] = relns

                else:
                    attr_stix_id = p.as_remote(r_tx).get_has(attribute_type=stix_id)
                    for attr in attr_stix_id:
                        play["stix_id"] = attr.get_value()

                role_i['player'].append(play)
            elif p.is_attribute():
                play["type"] = "attribute"
                play["tql"] = p.get_type().get_label().name()
                play["value"] = process_value(p)
                role_i['player'].append(play)

            else:
                logger.debug(f'player is not entity type {p}')

        roles.append(role_i)
    return roles


def validate_get_relns(rel,
                       r_tx,
                       obj_name,
                       import_type: ImportType):
    """
        When processing relations for an object, ensure we only access relations for sub objects,
        and not Stix relations or sightings
    Args:
        rel (): the relation
        r_tx (): the transaction
        obj_name (): the object involved in the relation

    Returns:
        reln {}: a dict containing the reln details
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    reln={}
    reln_name = rel.get_type().get_label().name()
    if reln_name in auth["tql_types"]["embedded_relations"]:
        for emb in auth["reln"]["embedded_relations"]:
            if emb['typeql'] == reln_name:
                role_owner = emb['owner']
        return return_valid_relations(rel, r_tx, obj_name, role_owner, import_type)

    elif reln_name in auth["tql_types"]["key_value_relations"]:
        for kvt in auth["reln"]["key_value_relations"]:
            if kvt['typeql'] == reln_name:
                role_owner = kvt['owner']
        return return_valid_relations(rel, r_tx, obj_name, role_owner, import_type)

    elif reln_name in auth["tql_types"]["extension_relations"]:
        logger.debug(f'reln name {reln_name}')
        for kvt in auth["reln"]["extension_relations"]:
            if kvt['relation'] == reln_name:
                role_owner = kvt['owner']
        return return_valid_relations(rel, r_tx, obj_name, role_owner, import_type)

    elif reln_name in auth["tql_types"]["list_of_objects"]:
        for kvt in auth["reln"]["list_of_objects"]:
            if kvt['typeql'] == reln_name:
                role_owner = kvt['owner']
        return return_valid_relations(rel, r_tx, obj_name, role_owner, import_type)

    elif reln_name == "granular-marking":
        return get_relation_details(rel, r_tx, import_type)

    elif reln_name == "hashes":
        return get_relation_details(rel, r_tx, import_type)

    else:
        logger.error(f'Error, relation name is {reln_name}')


def return_valid_relations(rel,
                           r_tx,
                           obj_name,
                           role_owner,
                           import_type: ImportType):
    """
        return only the valid relations to the relation check
    Args:
        rel (): the actual relation
        r_tx (): the transaction
        obj_name (): the object involved
        role_owner (): the owner of the role

    Returns:
        reln {}: a dict containing the reln details
    """
    reln_map = rel.as_remote(r_tx).get_players_by_role_type()
    for role, player in reln_map.items():
        role_name = role.get_label().name()
        if role_name == role_owner:
            for p in player:
                if p.is_entity():
                    play_name = p.get_type().get_label().name()
                    if play_name == obj_name:
                        return get_relation_details(rel, r_tx, import_type)


def get_standard_relations(r, r_tx):
    """
        Ignore standard relations, as they are not sub objects
    Returns:
        Emptyy List:
    """
    return []



"""Python Mitre ATT&CK Stix Class Definitions """
import json
import pathlib
from collections import OrderedDict

from stix2.exceptions import (
    PropertyPresenceError, )
from stix2.properties import (
    BooleanProperty, ExtensionsProperty, IDProperty, IntegerProperty, ListProperty,
    OpenVocabProperty, ReferenceProperty, StringProperty,
    TimestampProperty, TypeProperty, EmbeddedObjectProperty
)
from stix2.utils import NOW, _get_dict
from stix2.markings import _MarkingsMixin
from stix2.markings.utils import check_tlp_marking
from stix2.v21.base import _DomainObject, _STIXBase21, _RelationshipObject
from stix2.v21.common import (
    ExternalReference, GranularMarking, KillChainPhase,
    MarkingProperty, TLPMarking, StatementMarking,
)
from stix2.v21.vocab import (
    ATTACK_MOTIVATION, ATTACK_RESOURCE_LEVEL, IMPLEMENTATION_LANGUAGE, MALWARE_CAPABILITIES, MALWARE_TYPE,
    PROCESSOR_ARCHITECTURE, TOOL_TYPE, IDENTITY_CLASS, INDUSTRY_SECTOR,
)

import logging

from stixorm.module.definitions.property_definitions import ThreatReference
from stixorm.module.typedb_lib.auth_types import all_auth_types
from stixorm.module.typedb_lib.factories.mappings_factory import get_mapping_factory_instance

logger = logging.getLogger(__name__)

valid_obj =  get_mapping_factory_instance().get_all_types()
# i) allows x- prefix see properties.py line 592 obj_type.startswith("x-") and
# ii) allows non stix definitiosn to be added see properties.py line 592 is_object(obj_type, self.spec_version)

class AttackRelation(_RelationshipObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'relationship'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('relationship_type', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('source_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1', required=True)),
        ('target_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1', required=True)),
        ('start_time', TimestampProperty()),
        ('stop_time', TimestampProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

    # Explicitly define the first three kwargs to make readable Relationship declarations.
    # def __init__(
    #     self, source_ref=None, relationship_type=None,
    #     target_ref=None, **kwargs
    # ):
    #     # Allow (source_ref, relationship_type, target_ref) as positional args.
    #     if source_ref and not kwargs.get('source_ref'):
    #         kwargs['source_ref'] = source_ref
    #     if relationship_type and not kwargs.get('relationship_type'):
    #         kwargs['relationship_type'] = relationship_type
    #     if target_ref and not kwargs.get('target_ref'):
    #         kwargs['target_ref'] = target_ref
    #
    #     super(AttackRelation, self).__init__(**kwargs)

    def _check_object_constraints(self):
        super(self.__class__, self)._check_object_constraints()

        start_time = self.get('start_time')
        stop_time = self.get('stop_time')

        if start_time and stop_time and stop_time <= start_time:
            msg = "{0.id} 'stop_time' must be later than 'start_time'"
            raise ValueError(msg.format(self))

class Matrix(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'x-mitre-matrix'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('tactic_refs', ListProperty(StringProperty)),
    ])


class Tactic(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'x-mitre-tactic'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_shortname', StringProperty()),
    ])

class Technique(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'attack-pattern'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('x_mitre_network_requirements', BooleanProperty(default=lambda: False)),
        ('aliases', ListProperty(StringProperty)),
        ('kill_chain_phases', ListProperty(KillChainPhase)),
        ('x_mitre_network_requirements', BooleanProperty(default=lambda: False)),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_detection', StringProperty()),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('x_mitre_data_sources', ListProperty(StringProperty)),
        ('x_mitre_is_subtechnique', BooleanProperty(default=lambda: False)),
        ('x_mitre_system_requirements', ListProperty(StringProperty)),
        ('x_mitre_tactic_type', ListProperty(StringProperty)),
        ('x_mitre_permissions_required', ListProperty(StringProperty)),
        ('x_mitre_effective_permissions', ListProperty(StringProperty)),
        ('x_mitre_defense_bypassed', ListProperty(StringProperty)),
        ('x_mitre_remote_support', BooleanProperty(default=lambda: False)),
        ('x_mitre_impact_type', ListProperty(StringProperty)),
    ])


class SubTechnique(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'attack-pattern'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('aliases', ListProperty(StringProperty)),
        ('kill_chain_phases', ListProperty(KillChainPhase)),
        ('x_mitre_network_requirements', BooleanProperty(default=lambda: False)),
        ('revoked', BooleanProperty(default=lambda: True)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_detection', StringProperty()),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('x_mitre_data_sources', ListProperty(StringProperty)),
        ('x_mitre_is_subtechnique', BooleanProperty(default=lambda: True)),
        ('x_mitre_system_requirements', ListProperty(StringProperty)),
        ('x_mitre_tactic_type', ListProperty(StringProperty)),
        ('x_mitre_permissions_required', ListProperty(StringProperty)),
        ('x_mitre_effective_permissions', ListProperty(StringProperty)),
        ('x_mitre_defense_bypassed', ListProperty(StringProperty)),
        ('x_mitre_remote_support', BooleanProperty(default=lambda: False)),
        ('x_mitre_impact_type', ListProperty(StringProperty)),
    ])


class Mitigation(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'course-of-action'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

class Group(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'intrusion-set'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('aliases', ListProperty(StringProperty)),
        ('first_seen', TimestampProperty()),
        ('last_seen', TimestampProperty()),
        ('goals', ListProperty(StringProperty)),
        ('resource_level', OpenVocabProperty(ATTACK_RESOURCE_LEVEL)),
        ('primary_motivation', OpenVocabProperty(ATTACK_MOTIVATION)),
        ('secondary_motivations', ListProperty(OpenVocabProperty(ATTACK_MOTIVATION))),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

    def _check_object_constraints(self):
        super(Group, self)._check_object_constraints()

        first_seen = self.get('first_seen')
        last_seen = self.get('last_seen')

        if first_seen and last_seen and last_seen < first_seen:
            msg = "{0.id} 'last_seen' must be greater than or equal to 'first_seen'"
            raise ValueError(msg.format(self))


class SoftwareMalware(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'malware'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty()),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('malware_types', ListProperty(OpenVocabProperty(MALWARE_TYPE))),
        ('is_family', BooleanProperty(required=True)),
        ('aliases', ListProperty(StringProperty)),
        ('kill_chain_phases', ListProperty(KillChainPhase)),
        ('first_seen', TimestampProperty()),
        ('last_seen', TimestampProperty()),
        ('operating_system_refs', ListProperty(ReferenceProperty(valid_types='software', spec_version='2.1'))),
        ('architecture_execution_envs', ListProperty(OpenVocabProperty(PROCESSOR_ARCHITECTURE))),
        ('implementation_languages', ListProperty(OpenVocabProperty(IMPLEMENTATION_LANGUAGE))),
        ('capabilities', ListProperty(OpenVocabProperty(MALWARE_CAPABILITIES))),
        ('sample_refs', ListProperty(ReferenceProperty(valid_types=['artifact', 'file'], spec_version='2.1'))),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('x_mitre_aliases', ListProperty(StringProperty)),
    ])

    def _check_object_constraints(self):
        super(SoftwareMalware, self)._check_object_constraints()

        first_seen = self.get('first_seen')
        last_seen = self.get('last_seen')

        if first_seen and last_seen and last_seen < first_seen:
            msg = "{0.id} 'last_seen' must be greater than or equal to 'first_seen'"
            raise ValueError(msg.format(self))

        if self.is_family and "name" not in self:
            raise PropertyPresenceError(
                "'name' is a required property for malware families",
                SoftwareMalware,
            )


class SoftwareTool(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'tool'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('tool_types', ListProperty(OpenVocabProperty(TOOL_TYPE))),
        ('aliases', ListProperty(StringProperty)),
        ('kill_chain_phases', ListProperty(KillChainPhase)),
        ('tool_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('x_mitre_aliases', ListProperty(StringProperty)),
    ])



class DataSource(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'x-mitre-data-source'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_platforms', ListProperty(StringProperty)),
        ('x_mitre_collection_layers', ListProperty(StringProperty)),
    ])


class DataComponent(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'x-mitre-data-component'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_data_source_ref', StringProperty())
    ])

class AttackCampaign(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'campaign'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('x_mitre_first_seen_citation', StringProperty()),
        ('x_mitre_last_seen_citation', StringProperty()),
        ('x_mitre_aliases', StringProperty()),
        ('aliases', ListProperty(StringProperty)),
        ('first_seen', TimestampProperty()),
        ('last_seen', TimestampProperty()),
        ('objective', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

    def _check_object_constraints(self):
        super(AttackCampaign, self)._check_object_constraints()

        first_seen = self.get('first_seen')
        last_seen = self.get('last_seen')

        if first_seen and last_seen and last_seen < first_seen:
            msg = "{0.id} 'last_seen' must be greater than or equal to 'first_seen'"
            raise ValueError(msg.format(self))



class ObjectVersion(_STIXBase21):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """
    _properties = OrderedDict([
        ('object_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1')),
        ('object_modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
    ])


class Collection(_DomainObject):
    """For more detailed information on this object's properties, see
    `the MITRE ATT&CK Stix specifications <https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md>`__.
    """

    _type = 'x-mitre-collection'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('x_mitre_deprecated', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('x_mitre_contents', ListProperty(EmbeddedObjectProperty(type=ObjectVersion))),
    ])

class AttackMarking(_STIXBase21, _MarkingsMixin):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_k5fndj2c7c1k>`__.
    """

    _type = 'marking-definition'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('definition_type', StringProperty()),
        ('name', StringProperty()),
        ('definition', MarkingProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

    def __init__(self, **kwargs):
        if {'definition_type', 'definition'}.issubset(kwargs.keys()):
            # Create correct marking type object
            try:
                marking_type = OBJ_MAP_MARKING[kwargs['definition_type']]
            except KeyError:
                raise ValueError("definition_type must be a valid marking type")

            if not isinstance(kwargs['definition'], marking_type): # noqa
                defn = _get_dict(kwargs['definition'])
                kwargs['definition'] = marking_type(**defn)

        super(AttackMarking, self).__init__(**kwargs)

    def _check_object_constraints(self):
        super(AttackMarking, self)._check_object_constraints()

        definition = self.get("definition")
        definition_type = self.get("definition_type")
        extensions = self.get("extensions")

        if not (definition_type and definition) and not extensions:
            raise PropertyPresenceError(
                "MarkingDefinition objects must have the properties "
                "'definition_type' and 'definition' if 'extensions' is not present",
                AttackMarking,
            )

        check_tlp_marking(self, '2.1')

    def serialize(self, pretty=False, include_optional_defaults=False, **kwargs):
        check_tlp_marking(self, '2.1')
        return super(AttackMarking, self).serialize(pretty, include_optional_defaults, **kwargs)


OBJ_MAP_MARKING = {
    'tlp': TLPMarking,
    'statement': StatementMarking,
}

class AttackIdentity(_DomainObject):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_wh296fiwpklp>`__.
    """

    _type = 'identity'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('x_mitre_version', StringProperty()),
        ('x_mitre_contributors', ListProperty(StringProperty)),
        ('x_mitre_modified_by_ref', StringProperty()),
        ('x_mitre_domains', ListProperty(StringProperty)),
        ('x_mitre_attack_spec_version', StringProperty()),
        ('roles', ListProperty(StringProperty)),
        ('identity_class', OpenVocabProperty(IDENTITY_CLASS)),
        ('sectors', ListProperty(OpenVocabProperty(INDUSTRY_SECTOR))),
        ('contact_information', StringProperty()),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])


#
# Copyright (C) 2022 Vaticle
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os

from typedb.client import *
import logging

logger = logging.getLogger(__name__)

attack_raw = "https://raw.githubusercontent.com/mitre-attack/attack-stix-data/master/index.json"

# make sure the four TLP Markings are loaded when the database initialises
initial_markings = [[
    '$mark isa tlp-white, has stix-type "marking-definition"',
    ', has stix-id "marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9"',
    ', has spec-version "2.1", has created 2017-01-20T00:00:00.000;'
], [
    '$mark isa tlp-green, has stix-type "marking-definition"',
    ', has stix-id "marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da"',
    ', has spec-version "2.1", has created 2017-01-20T00:00:00.000;'
], [
    '$mark isa tlp-amber, has stix-type "marking-definition"',
    ', has stix-id "marking-definition--f88d31f6-486f-44da-b317-01333bde0b82"',
    ', has spec-version "2.1", has created 2017-01-20T00:00:00.000;'
], [
    '$mark isa tlp-red, has stix-type "marking-definition"',
    ', has stix-id "marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed"',
    ', has spec-version "2.1", has created 2017-01-20T00:00:00.000;'
]]
tlp_ids = ["marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9",
                   "marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da",
                   "marking-definition--f88d31f6-486f-44da-b317-01333bde0b82",
                   "marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed"]

def setup_database(stix_connection: Dict[str, str], clear: bool):
    url = stix_connection["uri"] + ":" + stix_connection["port"]
    with TypeDB.core_client(url) as client:
        logger.debug(f'Database Clearing is [{clear}]')
        if client.databases().contains(stix_connection["database"]):
            if clear:
                client.databases().get(stix_connection["database"]).delete()
                client.databases().create(stix_connection["database"])
            else:
                return
                # raise ValueError(f"Database '{database}' already exists")
        else:
            client.databases().create(stix_connection["database"])

        logger.debug('.......................... clear complete')


def load_schema(stix_connection: Dict[str, str], rel_path=None, schema_type: str = "schema"):
    logger.debug(f'{stix_connection}')
    logger.debug(rel_path)
    logger.debug(schema_type)
    assert rel_path is not None, "Need a path to history a schema"
    assert os.path.exists(rel_path), "File path needs to exist"

    url = stix_connection["uri"] + ":" + stix_connection["port"]
    with TypeDB.core_client(url) as client:
        # Stage 1: Create the schema
        with client.session(stix_connection["database"], SessionType.SCHEMA) as session:
            # Load schema from file
            with open(rel_path, "r") as schema_file:
                schema = schema_file.read()
            logger.debug('.....')
            logger.debug(f'Inserting {schema_type} ...')
            logger.debug('.....')
            with session.transaction(TransactionType.WRITE) as write_transaction:
                write_transaction.query().define(schema)
                write_transaction.commit()
            logger.debug('.....')
            logger.debug('Successfully committed schema!')
            logger.debug('.....')
            session.close()


def load_markings(stix_connection: Dict[str, str]):
    type_ql_list = []
    logger.info(f"========================== Database initialisation ============================")
    for mark_list in initial_markings:
        type_ql = " insert "
        for line in mark_list:
            type_ql += line
        type_ql_list.append(type_ql)
    load_typeql_data(type_ql_list, stix_connection)
    logger.info(f"===============================================================================\n\n")
    return_list = tlp_ids
    return return_list


def load_typeql_data(data_list, stix_connection: Dict[str, str]):
    url = stix_connection["uri"] + ":" + stix_connection["port"]
    with TypeDB.core_client(url) as client:
        # Stage 1: Create the schema
        with client.session(stix_connection["database"], SessionType.DATA) as session:
            with session.transaction(TransactionType.WRITE) as write_transaction:
                logger.debug(f'Loading TLP markings')
                for data in data_list:
                    logger.debug(f'\n\n{data}\n\n')
                    insert_iterator = write_transaction.query().insert(data)

                    logger.debug(f'insert_iterator response ->\n{insert_iterator}')
                    for result in insert_iterator:
                        logger.info(f'typedb response ->\n{result}')

                write_transaction.commit()




def sort_layers(layers,
                cyclical,
                indexes: List[str],
                missing: List[str],
                dep_obj, add_or_del='del'):
    """ Sort the layers depending on whether they are "add" or "del" layers

    Args:
        layers ():
        cyclical ():
        indexes ():
        missing ():
        dep_obj ():
        add_or_del ():

    Returns:

    """
    logger.debug(
        f"################################### enter sort_layers {add_or_del} ###############################################")
    logger.debug(f'\nlayers -> {layers}\ncyclical indexes -> {cyclical}\nindexes -> {indexes}\nmissing -> {missing}')
    logger.debug(f'add_or_del -> {add_or_del}\ndep_obj -> {dep_obj}')
    logger.debug("-------------------------------  ------------------------------------------------")
    # Stage 1 - Initialise Variables
    # 1. Setup key variables
    loc_id = dep_obj['id']
    loc_list = dep_obj["dep_list"]
    locset = set(loc_list)
    mset = set(missing)
    iset = set(indexes)
    intset = locset.intersection(iset)
    diffset = locset.difference(intset)
    # Stage 2 - Analyse Choices
    # 2. check whether the object has dependencies in its dep_list
    if not loc_list:
        # 2.a empty depedency list
        dep_list_items = False
        logger.debug('### There are no dependencies')
    else:
        # 2.b dependency list has indexes in it
        dep_list_items = True
        logger.debug('### There are some dependencies')
    # 3. check whether the object id is in the missing list
    if loc_id in mset:
        id_in_missing = True
        logger.debug('### Current is the missing dependency of an existing record')
    else:
        id_in_missing = False
        logger.debug('### Current is not a missing dependency of an existing record')
    # 4. check whether any id in the object dependency list is not already loaded
    if not diffset:
        dep_id_not_loaded = False
        logger.debug('### No dependencies need to be added to missing')
    else:
        dep_id_not_loaded = True
        mset = mset | diffset
        logger.debug('### Missing - Some dependencies need to be added to missing')

    # Stage 3 - Execute Choices
    # 5. No dependencies and not in missing, append only
    if not dep_list_items and not id_in_missing:
        logger.debug('### Append object and return')
        if add_or_del == 'del':
            layers.append(dep_obj)
            indexes.append(loc_id)
        elif add_or_del == 'add':
            layers.insert(0, dep_obj)
            indexes.insert(0, loc_id)
        logger.debug(f'layers -> {layers}\nindexes -> {indexes}\nmset -> {mset}')
        logger.debug(
            f"################################## end of  sort_layers {add_or_del} ####################################################")
        return layers, indexes, list(mset), cyclical
    # 6. There are no dependencies but id is in missing, delete from missing,follow the tree and reorder
    if not dep_list_items and id_in_missing:
        logger.debug('### delete from missing,follow the tree and reorder')
        mset.remove(loc_id)
        tree, circular = follow_the_tree(layers, dep_obj)
        cyclical = cyclical + circular
        logger.debug(f' tree -> {tree}')
        layers, indexes = reorder(layers, indexes, tree, dep_obj, add_or_del)
        logger.debug(f'layers -> {layers}\nindexes -> {indexes}\nmset -> {mset}')
        logger.debug(
            f"################################## end of  sort_layers {add_or_del} ####################################################")
        return layers, indexes, list(mset), cyclical
    # 7 There are dependencies, object is not in missing, insert at front
    if dep_list_items and not id_in_missing:
        logger.debug('### Add current to the front of the record')
        if add_or_del == 'del':
            layers.insert(0, dep_obj)
            indexes.insert(0, loc_id)
        elif add_or_del == 'add':
            layers.append(dep_obj)
            indexes.append(loc_id)
        logger.debug(f'layers -> {layers}\nindexes -> {indexes}\nmset -> {mset}')
        logger.debug(
            f"################################## end of  sort_layers {add_or_del} ####################################################")
        return layers, indexes, list(mset), cyclical
    # 8 There are dependencies, object is in missing , delete from missing,follow the tree and reorder
    if dep_list_items and id_in_missing:
        logger.debug(f'### delete from missing,follow the tree and reorder')
        mset.remove(loc_id)
        tree, circular = follow_the_tree(layers, dep_obj)
        cyclical = cyclical + circular
        logger.debug(f' tree -> {tree}')
        layers, indexes = reorder(layers, indexes, tree, dep_obj, add_or_del)
        logger.debug(f'layers -> {layers}\nindexes -> {indexes}\nmset -> {mset}')
        logger.debug(
            f"################################## end of  sort_layers {add_or_del} ####################################################")
        return layers, indexes, list(mset), cyclical

    logger.debug("theres a massive problem")
    return layers, indexes, list(mset), cyclical



def reorder(layers, indexes, tree, dep_obj, add_or_del):
    """ Reorder the layers list of objects based on dependencies
    Args:
        layers (): ordered list of layers
        indexes (): list of indexes in the same order as layers
        tree (): list of indexes, based on finding a hierarchy depdnent on this record
        dep_obj (): a dependency object containing the new layer element to be added
        add_or_del (): a flag on whether the records represent insert or delete operations (order)
    Returns:
        layers (): ordered list of layers now including the dep_object
        indexes (): ordered list of stix ids to match the layers
    """
    front_layers = []
    front_indexes = []
    dep_layers = []
    dep_indexes = []
    dtree = []
    loc_list = dep_obj["dep_list"]
    tree = list(set(tree))
    logger.debug("%%%%%%%%%%%%%%% reorder 1 %%%%%%%%%%%%%%%%%%")
    logger.debug(f'\n orig indexes -> {indexes}\n orig layers, {layers}\n dep_obj , {dep_obj}')
    logger.debug("%%%%%%%%%%%%%%% reorder 2 %%%%%%%%%%dep_obj%%%%%%%%")
    # 1. Copy elements from layers and indexes so they are in the order we want them
    if add_or_del == 'del':
        up_tree = reversed(tree)
    elif add_or_del == 'add':
        up_tree = tree
    for t in up_tree:
        front_layers.append(layers[t])
        front_indexes.append(indexes[t])
    # 1.B Handle the dep_list linkages
    for d in loc_list:
        for index, i in enumerate(indexes):
            if d == i:
                dtree.append(index)
                dep_layers.append(layers[index])
                dep_indexes.append(layers[index]['id'])
    # 2. Now order the tree in reverse numeric order, biggest first
    total_tree = tree + dtree
    total_tree.sort(reverse=True)
    # 3. Now delete the elements from layers and indexes
    # TODO: Error here for test_add_files
    for t in total_tree:
        layers.pop(t)
        indexes.pop(t)
    # 4. Add the dep_obj to the new lists
    if add_or_del == 'del':
        front_indexes.append(dep_obj['id'])
        front_layers.append(dep_obj)
        layers = front_layers + dep_layers + layers
        indexes = front_indexes + dep_indexes + indexes
    elif add_or_del == 'add':
        front_layers.insert(0, dep_obj)
        front_indexes.insert(0, dep_obj['id'])
        layers = dep_layers + front_layers + layers
        indexes = dep_indexes + front_indexes + indexes
    # 5. Assemble the final lists
    logger.debug(
        f'\nfront_indexes -> {front_indexes}\n\nfront_layers, {front_layers}\n\n old layers, {layers}\n\nold indexes -> {indexes}')
    logger.debug("-------------------------------------------------------------------------------------")
    logger.debug("%%%%%%%%%%%%%%% end reorder %%%%%%%%%%%%%%%%%%")
    return layers, indexes


def follow_the_tree(layers, dep_obj):
    """ Follow the tree of dependencies and report back the list and any cyclical id's

    Args:
        layers ():
        dep_obj ():

    Returns:
        tree: a ist of imdexes numbers
        cyclical: a list of stix-ds in a cyclical relationship (error condition)

    """
    tree = []
    cyclical = []
    loc_id = dep_obj['id']
    loc_ids = [loc_id]
    loc_dep = dep_obj['dep_list']
    found = True
    found, ret_indexes, ret_ids, circular = find_id(loc_ids, loc_dep, layers)
    logger.debug(f'### Following First Level Tree, found {found},index {ret_indexes}, loc_id {ret_ids}')
    if found:
        tree = tree + ret_indexes
        loc_ids = ret_ids
        cyclical = circular
    while found:
        found, ret_indexes, ret_ids, circular = find_id(loc_ids, loc_dep, layers)
        logger.debug(f'### Following Tree, found {found},index {ret_indexes}, loc_id {ret_ids}')
        if found:
            tree = tree + ret_indexes
            loc_ids = ret_ids
    return tree, cyclical


def find_id(loc_ids, loc_dep, layers):
    """ Compare a list of ids, and find which layers are dependent on this list

    Args:
        loc_ids (list): one or more stix ids describing the record
        loc_dep (list): a list of stix ids that this record is dependent on

    Returns:
        found (boolean): whether the stix-ids were found in the depednency's of the layers
        indexes (list): a list of the index numbers that are dependencies
        ids (list): a list of layer stix-ids that are dependent on this record
    """
    cyclical = []
    found = False
    indexes = []
    ids = []
    for loc_id in loc_ids:
        for index, lay in enumerate(layers):
            if loc_id in lay['dep_list']:
                if lay['id'] in loc_dep:
                    cyclical.append(lay['id'])
                    cyclical.append(loc_id)
                found = True
                indexes.append(index)
                ids.append(lay['id'])

    return found, indexes, ids, cyclical


# if this file is run directly, then start here
if __name__ == '__main__':
    # define the localhost and default stix2 setup
    connection = {
        "uri": "localhost",
        "port": "1729",
        "database": "stix2",
        "user": None,
        "password": None,
        "clear": True
    }

    import_type = {
        "STIX21": True,
        "CVE": False,
        "identity": False,
        "location": False,
        "rules": False,
        "ATT&CK": True,
        "ATT&CK_Versions": ["12.0"],
        "ATT&CK_Domains": ["enterprise-attack", "mobile-attack", "ics-attack"],
        "CACAO": False
    }



"""Python Mitre ATT&CK Stix Class Definitions """
from stixorm.module.typedb_lib.factories.mappings_factory import get_mapping_factory_instance

"""Python Mitre ATT&CK Stix Class Definitions """
import json
import pathlib
from collections import OrderedDict

from stix2.exceptions import (
    PropertyPresenceError, )
from stix2.properties import (
    BooleanProperty, ExtensionsProperty, IDProperty, IntegerProperty, ListProperty,
    OpenVocabProperty, ReferenceProperty, StringProperty, FloatProperty,
    TimestampProperty, TypeProperty, EmbeddedObjectProperty
)
from stix2.utils import NOW, _get_dict
from stix2.markings import _MarkingsMixin
from stix2.markings.utils import check_tlp_marking
from stix2.v21.base import _DomainObject, _STIXBase21, _RelationshipObject, _Extension
from stix2.v21.common import (
    ExternalReference, GranularMarking, KillChainPhase,
    MarkingProperty, TLPMarking, StatementMarking,
)
from stix2.v21.vocab import (
    ATTACK_MOTIVATION, ATTACK_RESOURCE_LEVEL, IMPLEMENTATION_LANGUAGE, MALWARE_CAPABILITIES, MALWARE_TYPE,
    PROCESSOR_ARCHITECTURE, TOOL_TYPE, IDENTITY_CLASS, INDUSTRY_SECTOR,
)

import logging

from stixorm.module.definitions.property_definitions import ThreatReference, ThreatExtensionsProperty


logger = logging.getLogger(__name__)

valid_obj =  get_mapping_factory_instance().get_all_types()
# i) allows x- prefix see properties.py line 592 obj_type.startswith("x-") and
# ii) allows non stix definitiosn to be added see properties.py line 592 is_object(obj_type, self.spec_version)

############################################################################################
#
# Feed Definitions
#
############################################################################################
class ThreatSubObject(_STIXBase21):
    """For more detailed information on this object's properties, see
    `the OS-Threat documentation`__.
    """
    _properties = OrderedDict([
        ('object_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
    ])


class Feed(_DomainObject):
    """For more detailed information on this object's properties, see
    `the OS-Threat documentation`__.
    """
    _type = 'feed'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('paid', BooleanProperty()),
        ('free', BooleanProperty()),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('contents', ListProperty(EmbeddedObjectProperty(type=ThreatSubObject))),
    ])



class Feeds(_DomainObject):
    """For more detailed information on this object's properties, see
    `the OS-Threat documentation`__.
    """
    _type = 'feeds'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('paid', BooleanProperty()),
        ('free', BooleanProperty()),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
        ('contained', ListProperty(ThreatReference(valid_types='feed', spec_version='2.1'))),
    ])


############################################################################################
#
# Incident Definitions
#
############################################################################################
# Event Object
##################################################################################
class StateChangeObject(_STIXBase21):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _properties = OrderedDict([
        ('state_change_type', StringProperty()),
        ('initial_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1')),
        ('result_ref', ThreatReference(valid_types=valid_obj, spec_version='2.1')),
    ])


class EventCoreExt(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """

    _type = 'extension-definition--4ca6de00-5b0d-45ef-a1dc-ea7279ea910e'
    _properties = OrderedDict([
        ('extension_type', StringProperty(fixed='new-sdo')),
    ])


class Event(_DomainObject):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _type = 'event'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('status', StringProperty()),
        ('changed_objects', ListProperty(EmbeddedObjectProperty(type=StateChangeObject))),
        ('description', StringProperty()),
        ('detection_methods', ListProperty(StringProperty)),
        ('detection_rule', StringProperty()),
        ('detection_system', StringProperty()),
        ('end_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('end_time_fidelity', StringProperty()),
        ('event_seq', IntegerProperty()),
        ('event_types', ListProperty(StringProperty)),
        ('goal', StringProperty()),
        ('name', StringProperty()),
        ('sighting_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'))),
        ('start_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('start_time_fidelity', StringProperty()),
        ('subevents', ListProperty(ThreatReference(valid_types='event', spec_version='2.1'))),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ThreatExtensionsProperty(spec_version='2.1')),
    ])
###############################################################################
# Imapct object
#################################################################################


class EntityCountObject(_STIXBase21):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _properties = OrderedDict([
        ('individual', IntegerProperty()),
        ('group', IntegerProperty()),
        ('system', IntegerProperty()),
        ('organization', IntegerProperty()),
        ('class', IntegerProperty()),
        ('unknown', IntegerProperty()),
    ])


class ImpactCoreExt(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """

    _type = 'extension-definition--7cc33dd6-f6a1-489b-98ea-522d351d71b9'
    _properties = OrderedDict([
        ('extension_type', StringProperty(fixed='new-sdo')),
    ])


class Availability(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'availability'
    _properties = OrderedDict([
        ('availability_impact', IntegerProperty()),
    ])


class Confidentiality(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'confidentiality'
    _properties = OrderedDict([
        ('information_type', StringProperty()),
        ('loss_type', StringProperty()),
        ('record_count', IntegerProperty()),
        ('record_size', IntegerProperty()),
    ])

class External(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'external'
    _properties = OrderedDict([
        ('impact_type', StringProperty()),
    ])


class Integrity(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'integrity'
    _properties = OrderedDict([
        ('alteration', StringProperty()),
        ('information_type', StringProperty()),
        ('record_count', IntegerProperty()),
        ('record_size', IntegerProperty()),
    ])


class Monetary(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'monetary'
    _properties = OrderedDict([
        ('variety', StringProperty()),
        ('conversion_rate', FloatProperty()),
        ('conversion_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('currency', StringProperty()),
        ('currency_actual', StringProperty()),
        ('max_amount', FloatProperty()),
        ('min_amount', FloatProperty()),
    ])


class Physical(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'physical'
    _properties = OrderedDict([
        ('impact_type', StringProperty()),
        ('asset_type', StringProperty()),
    ])



class Traceability(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/dod-cyber-crime-center/cti-stix-common-objects/blob/incident_rework/extension-definition-specifications/incident-core/Incident%20Extension%20Suite.adoc`__.
    """

    _type = 'traceability'
    _properties = OrderedDict([
        ('traceability_impact', StringProperty()),
    ])


class Impact(_DomainObject):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _type = 'impact'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('impact_category', StringProperty()),
        ('criticality', IntegerProperty()),
        ('description', StringProperty()),
        ('end_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('end_time_fidelity', StringProperty()),
        ('impacted_entity_counts', EmbeddedObjectProperty(type=EntityCountObject)),
        ('impacted_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'))),
        ('recoverability', StringProperty()),
        ('start_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('start_time_fidelity', StringProperty()),
        ('superseded_by_ref', ThreatReference(valid_types='impact', spec_version='2.1')),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ThreatExtensionsProperty(spec_version='2.1')),
    ])


###############################################################################
# Incident Extension object
#################################################################################

class IncidentScoreObject(_STIXBase21):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _properties = OrderedDict([
        ('name', StringProperty()),
        ('value', IntegerProperty()),
        ('description', StringProperty()),
    ])


class IncidentCoreExt(_Extension):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """

    _type = 'extension-definition--ef765651-680c-498d-9894-99799f2fa126'
    _properties = OrderedDict([
        ('determination', StringProperty()),
        ('extension_type', StringProperty(fixed='property-extension')),
        ('investigation_status', StringProperty()),
        ('criticality', IntegerProperty(min=0)),
        ('blocked', BooleanProperty()),
        ('malicious', BooleanProperty()),
        ('impacted_entity_counts', EmbeddedObjectProperty(type=EntityCountObject)),
        ('recoverability', ListProperty(StringProperty)),
        ('scores', EmbeddedObjectProperty(type=IncidentScoreObject)),
        ('incident_types', ListProperty(StringProperty)),
        ('task_refs', ListProperty(ThreatReference(valid_types='task'))),
        ('event_refs', ListProperty(ThreatReference(valid_types='event'))),
        ('impact_refs', ListProperty(ThreatReference(valid_types='impact'))),
        ('notes_refs', ListProperty(ThreatReference(valid_types='notes'))),
        ('evidence_refs', ListProperty(ThreatReference(valid_types='evidence'))),
    ])

###############################################################################
# Task object
#################################################################################

class TaskCoreExt(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """

    _type = 'extension-definition--2074a052-8be4-4932-849e-f5e7798e0030'
    _properties = OrderedDict([
        ('extension_type', StringProperty(fixed='new-sdo')),
    ])


class Task(_DomainObject):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _type = 'task'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('changed_objects', ListProperty(EmbeddedObjectProperty(type=StateChangeObject))),
        ('task_type', StringProperty()),
        ('step_type', StringProperty()),
        ('outcome', StringProperty()),
        ('description', StringProperty()),
        ('end_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('end_time_fidelity', StringProperty()),
        ('error', StringProperty()),
        ('impacted_entity_counts', EmbeddedObjectProperty(type=EntityCountObject)),
        ('name', StringProperty(required=True)),
        ('priority', IntegerProperty(min=0)),
        ('start_time', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('start_time_fidelity', StringProperty()),
        ('owner', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('on_completion', ThreatReference(valid_types='task')),
        ('on_failure', ThreatReference(valid_types='task')),
        ('on_success', ThreatReference(valid_types='task')),
        ('next_steps', ListProperty(ThreatReference(valid_types='task'))),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])


###############################################################################
# Evidence object
#################################################################################

class EvidenceCoreExt(_Extension):
    """For more detailed information on this object's properties, see
    `the  https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """

    _type = 'extension-definition--7ff5b5a5-a342-417e-9c0d-339561d9d78a'
    _properties = OrderedDict([
        ('extension_type', StringProperty(fixed='new-sdo')),
    ])


class Evidence(_DomainObject):
    """For more detailed information on this object's properties, see
    `the https://github.com/os-threat/stix-extensions/wiki/2.-Description-of-Incident-Model`__.
    """
    _type = 'evidence'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('evidence_type', StringProperty()),
        ('source', StringProperty()),
        ('object_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'))),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])




import traceback


import logging
logger = logging.getLogger(__name__)

def handle_result(result,
                  name: str = '',
                  strict_failure: bool = False):
        result_failure = result is not None
        try:
            if result_failure:
                logger.info("Failure in result for: "+ name)
                logging.exception("\n".join(traceback.format_exception(result.failure())))
        except Exception as e:
            logger.exception(e)
            raise Exception(str(result.failure()))

def handle_missing_values(result_failure: bool,
                  name: str = '',
                  strict_failure: bool = False):

    try:
        if result_failure:
            logger.error(name)
    except Exception as e:
        logger.exception(e)
        raise Exception(name)


import sys,os,json
import re

def check_status(key:str,status:dict):
    if key in status:
        return status[key]
    else:
        return False

def set_status(key:str,status:dict):

    status[key]=True

    for kk in status.keys():
        if key != kk: status[kk]=False


if __name__ == '__main__':

    with open('example.txt','r') as file:
        text = file.read()

        status = {}
        for no,line in enumerate(text.split('\n')):
            if line.startswith('core_ql'):
                print(f'Core found at {no}')
                # check rest of line
                query = line[line.index('>')+1:].strip()
                set_status('core_ql',status)

            elif check_status('core_ql',status):
                query = line.strip()

            if line.startswith('indep_ql'):
                print(f'Indep found at {no}')
                # check rest of line
                query = line[line.index('>')+1:].strip()
                if len(query)>0:
                    m = re.search(r"(\$[a-z\-\_]+)\s+isa\s+", query)
                    if m:
                        first_word = m.groups()[0]
                        print(f'Found first {first_word}')

                set_status('indep_ql',status)
            elif check_status('indep_ql',status):
                query = line.strip()
                if len(query)>0:
                    m = re.search(r"(\$[a-z\-\_]+)\s+isa\s+", query)
                    if m:
                        first_word = m.groups()[0]
                        print(f'Found first {first_word}')

            if line.startswith('dep_match'):
                print(f'Dep match found at {no}')
                # check rest of line
                query = line[line.index('>')+1:].strip()

                set_status('dep_match',status)
            elif check_status('dep_match',status):
                query = line.strip()

            if line.startswith('dep_insert ->'):
                print(f'Dep Insert found at {no}')
                # check rest of line
                query = line[line.index('>')+1:].strip()
                if len(query)>0:
                    m = re.search(r"[a-z\-\_]+;$",query)
                    last_word = m.group(0)
                    print(f'Found last {last_word}')

                set_status('dep_insert',status)
            elif check_status('dep_insert',status):
                query = line.strip()
                if len(query)>0:
                    m = re.search(r"[a-z\-\_]+;$",query)
                    if m:
                        last_word = m.group(0)
                        print(f'Found last {last_word}')










import csv
import logging
import os
import glob
import csv
import json

from stixorm.module.typedb_lib.factories.definition_factory import get_definition_factory_instance
from stixorm.module.typedb_lib.model.definitions import DefinitionName
logger = logging.getLogger(__name__)
stix_models = get_definition_factory_instance().lookup_definition(DefinitionName.STIX_21)
attack_models = get_definition_factory_instance().lookup_definition(DefinitionName.ATTACK)

heading_align = [
    ":--------------------|",
    ":--------------------------------:|",
    ":------------------:|",
    ":------------------------:|",
    ":-------------:|"
]

object_docs = [
    {
        "dir": "sdo",
        "protocol": "stix21",
        "file": "sdo.csv",
        "obj_type": "Domain"
    },
    {
        "dir": "sco",
        "protocol": "stix21",
        "file": "sco.csv",
        "obj_type": "Cyber Obervable"
    }
]


object_tables = [
    {
        "title": "OASIS Stix 2.1",
        "protocol": "stix21",
        "objects": stix_models.get_mapping("object_conversion")
    },
    {
        "title": "Mitre ATT&CK",
        "protocol": "attack",
        "objects": attack_models.get_mapping("object_conversion")
    }
]
protocols = [
    "stix21",
    "attack"
]

titles = {
    "stix21": "OASIS Stix 2.1",
    "attack": "MITRE ATT&CK",
    "os-threat": "OS-Threat Custom",
    "sdo": "Domain Object's Types",
    "sco": "Cyber Observable Object Types",
    "sro": "Relationship Object Types",
    "sub": "Sub-Object Types"
}
object_types = [
    "sdo",
    "sco",
    "sro",
    "sub"
]

def gen_tables(obj_tables):
    bucket = {}
    bucket["stix21"] = {}
    bucket["attack"] = {}
    bucket["stix21"]["sdo"] = []
    bucket["stix21"]["sro"] = []
    bucket["stix21"]["sco"] = []
    bucket["stix21"]["sub"] = []
    bucket["attack"]["sdo"] = []
    bucket["attack"]["sro"] = []
    bucket["attack"]["sco"] = []
    bucket["attack"]["sub"] = []
    for table in obj_tables:
        obj_set = table["objects"]
        protocol = table["protocol"]
        layer = {}
        for obj in obj_set:
            obj_type = obj["object"]
            bucket[protocol][obj_type].append(obj)

    return bucket


def print_normal_rows(rel_dir, size, outfile, bucket_list, icon_dir, md_dir):
    for obj in bucket_list:
        detail_string = ""
        icon = obj["icon"]
        name = obj["typeql"]
        docs = obj["doc_url"]
        summary = obj["summary"]
        #logger.info(f"---- summary -> {summary}")
        logger.info(f"icon dir -> {icon_dir}, md dir {md_dir}")
        if icon == "":
            icon_name = ""
            icon_dir_name = ""
        else:
            icon_name = name
            icon_dir_name = icon_dir + icon
        if docs == "":
            name_dir = ""
        else:
            name_dir = md_dir + docs

        icon_md = "![" + icon_name + "](" + icon_dir_name + ")"
        name_md = "[" + name + "](" + name_dir + ")"
        detail_string += "| " + icon_md + " | " + name_md + " | " + summary + " |"
        logger.info(detail_string, file=outfile)


def print_summary_rows(rel_dir, size, outfile, bucket_list, icon_dir, md_dir):
    num_objs = len(bucket_list)
    rows = num_objs // size
    remainder = num_objs % size
    rel_path = "protocols/"
    k = 0
    if remainder != 0:
        rows += 1
        filler = size - remainder
    for j in range(rows):
        detail_string = ""
        for i in range(size):
            icon = ""
            name = ""
            docs = ""
            if k < num_objs:
                layer = bucket_list[k]
                icon = layer["icon"]
                name = layer["typeql"]
                docs = layer["doc_url"]

            if icon == "":
                icon_name = ""
                icon_dir_name = ""
            else:
                icon_name = name
                icon_dir_name = icon_dir + icon
            if docs == "":
                name_dir = ""
            else:
                name_dir = md_dir + docs
            icon_md = "![" + icon_name + "](" + icon_dir_name + ")"
            name_md = "[" + name + "](" + name_dir + ")"
            detail_string += "| " + icon_md + " | " + name_md
            k += 1

        detail_string += " |"
        logger.info(detail_string, file=outfile)


def print_summary_tables(rel_dir, outfile, protocol, bucket, size):
    logger.info("In summary tables")
    logger.info(f"reldir 1-> {rel_dir}")

    # Setup Table Headers
    if rel_dir == "./docs":
        rel_dir = "./protocols"
    elif rel_dir == "./docs/protocols":
        rel_dir = "."
    elif rel_dir == "./docs/protocols/stix21":
        rel_dir = "."
    sub_head = ""
    logger.info(f"reldir 2-> {rel_dir}")
    head_string = ""
    under_string = ""
    row_string = ""
    for i in range(size):
        head_string += "| Icon | Object Type "
        under_string += "|:----------:|:-----------"
    if size == 1:
        head_string = "| Icon | Object Type | Description"
        under_string = "|:----------:|:-----------|:-----------"
    head_string += " |"
    under_string += " |"
    logger.info(f"I am ready to process summary tables, size {size}")
    logger.info(f"object types {object_types}")
    for obj_type in object_types:
        logger.info(f"\n###  {titles[obj_type]}\n", file=outfile)
        logger.info(f'obj type {obj_type}, protocol {protocol}, protocols {protocols}')
        if protocol == "all":
            for proto in protocols:
                local_bucket = bucket[proto][obj_type]
                logger.info(f'local bucket length is -> {len(local_bucket)}')
                if len(local_bucket) != 0:
                    md_dir = rel_dir + "/" + proto + "/" + obj_type + "/"
                    icon_dir = rel_dir + "/" + proto + "/icons/"
                    logger.info(f"#### {titles[proto]} \n", file=outfile)
                    logger.info(head_string, file=outfile)
                    logger.info(under_string, file=outfile)
                    if size == 1:
                        logger.info("go down the size==1 route and protocol !=1")
                        print_normal_rows(rel_dir, size, outfile, bucket[proto][obj_type], icon_dir, md_dir)
                    else:
                        logger.info("go down the size!=1 and protocol !=1 route")
                        print_summary_rows(rel_dir, size, outfile, bucket[proto][obj_type], icon_dir, md_dir)
                    logger.info("\n\n", file=outfile)
        else:
            local_bucket = bucket[protocol][obj_type]
            if len(local_bucket) != 0:
                md_dir = rel_dir + "/" + obj_type + "/"
                icon_dir = rel_dir + "/icons/"
                logger.info(f"#### {titles[protocol]} \n", file=outfile)
                logger.info(head_string, file=outfile)
                logger.info(under_string, file=outfile)
                if size == 1:
                    logger.info("go down the size==1 route")
                    print_normal_rows(rel_dir, size, outfile, bucket[protocol][obj_type], icon_dir, md_dir)
                else:
                    logger.info("go down the size!=1 route")
                    print_summary_rows(rel_dir, size, outfile, bucket[protocol][obj_type], icon_dir, md_dir)
                logger.info("\n\n", file=outfile)


def gen_overview_doc(rel_dir, bucket, protocol, size):
    logger.info("generate overview")
    # open up generic overview doc
    md_name = rel_dir + "/" + "overview.md"
    or_mname = rel_dir + "/" + "_orig.md"
    if os.path.exists(md_name):
        os.remove(md_name)
    outfile = open(md_name, "w")
    # insert the top part of the file
    origfile = open(or_mname, "r")
    lines = origfile.readlines()
    for line in lines:
        logger.info(line, file=outfile)
    # now generate overview table
    logger.info("## Total Objects in the System\n\n", file=outfile)
    try:
        print_summary_tables(rel_dir, outfile, protocol, bucket, size)
    except:
        logger.info("ERROR IN OBJECT Table GENERATION")
    # close the overview markdown file
    outfile.close()


def configure_overview_table_docs(obj_tables):
    logger.info("starting to process the list of objects")
    bucket = gen_tables(obj_tables)
    # Setup Library Overview Document
    rel_dir = "./docs"
    gen_overview_doc(rel_dir, bucket, "all", 3)
    # Setup Protocol Overview Document
    rel_dir = "./docs/protocols"
    gen_overview_doc(rel_dir, bucket, "all", 1)
    # Setup Library Overview Document
    rel_dir = "./docs/protocols/stix21"
    gen_overview_doc(rel_dir, bucket, "stix21", 1)



def delete_existing_markdown(dir):
    cwd = os.getcwd()
    del_dir = cwd + "\\" + dir
    del_pattern = del_dir + "\\" + "*.md"
    fileList = glob.glob(del_pattern)
    logger.info(f'delete fileList is {fileList}')
    for filePath in fileList:
        try:
            os.remove(filePath)
        except:
            logger.info("Error while deleting file : " ,filePath)
    return del_dir


def generate_object_doc(dir, fields, row, obj_type):
    image = row[0]
    table = row[1]
    stix_type = row[2]
    obj = row[3]
    para1 = row[4]
    para2 = row[5]
    url = row[6]
    json_example = row[7]
    tql_ins = row[8]
    tql_match = row[9]
    py_match = row[10]
    md_name = dir + "\\" + obj + ".md"
    outfile = open(md_name, "w")
    # 1. Setup Page title
    logger.info(f'# {obj} {obj_type} Object\n', file=outfile)
    logger.info(f'**Stix and TypeQL Object Type:**  `{stix_type}`\n', file=outfile)
    # 2. Setup Overview paragraphs
    logger.info(f'{para1}\n', file=outfile)
    if para2 != "":
        logger.info(f'{para2}\n', file=outfile)
    logger.info(f'[Reference in Stix2.1 Standard]({url})', file=outfile)
    # 3. Setup Table
    logger.info(f'## Stix 2.1 Properties Converted to TypeQL', file=outfile)
    logger.info(f'Mapping of the Stix Attack Pattern Properties to TypeDB\n', file=outfile)
    table_name = dir + "\\csv\\" + table
    rows = []
    with open(table_name, 'r') as csvfile:
        # creating a csv reader object
        csvreader = csv.reader(csvfile)

        # extracting field names through first row
        field_list = next(csvreader)
        fields = "| "
        for field in field_list:
            fields += field + " |"
        heading = "|"
        for head in heading_align:
            heading += head
        logger.info(f'{fields}', file=outfile)
        logger.info(f'{heading}', file=outfile)
        # extracting each data row one by one
        for row in csvreader:
            cols = "| "
            for col in row:
                cols += col + " |"
            logger.info(f'{cols}', file=outfile)
    # 4. Setup JSON Section
    logger.info(f'\n## The Example {obj} in JSON', file=outfile)
    logger.info(f'The original JSON, accessible in the Python environment', file=outfile)
    logger.info(f'```json\n{json_example}\n```\n', file=outfile)
    # 5. Setup TypeQL Insert Section
    logger.info(f'\n## Inserting the Example {obj} in TypeQL', file=outfile)
    logger.info(f'The TypeQL insert statement', file=outfile)
    logger.info(f'```typeql\n{tql_ins}\n```\n', file=outfile)
    # 6. Setup TypeQL Match Section
    logger.info(f'## Retrieving the Example {obj} in TypeQL', file=outfile)
    logger.info(f'The typeQL match statement\n', file=outfile)
    logger.info(f'```typeql\n{tql_match}\n```\n', file=outfile)
    # 7. Setup Force Graph Image Section
    logger.info(f'\nwill retrieve the example attack-pattern object in Vaticle Studio', file=outfile)
    imagefile = "./img/"+image
    logger.info(f'![{obj} Example]({imagefile})', file=outfile)
    # 8. Setup Python Match Section
    logger.info(f'\n## Retrieving the Example {obj}  in Python', file=outfile)
    logger.info(f'The Python retrieval statement\n', file=outfile)
    logger.info(f'```python\n{py_match}\n```\n', file=outfile)
    # 9. Close the File
    outfile.close()
    return


def gen_obj_docs(docs):
    # generate object docs
    for doc_set in docs:
        # Get details for each set of documents
        sub_dir = doc_set["dir"]
        doc_file = doc_set["file"]
        obj_type = doc_set["obj_type"]
        prot_type = doc_set["protocol"]
        # delete existing markdown documents
        rel_path = "protocols/" +prot_type +"/" + sub_dir
        full_doc_dir = delete_existing_markdown(rel_path)
        # generate new markdown docs in the target directory
        if full_doc_dir != "":
            file_path = full_doc_dir + "/" + doc_file
            rows=[]
            with open(file_path, 'r') as csvfile:
                # creating a csv reader object
                csvreader = csv.reader(csvfile)

                # extracting field names through first row
                fields = next(csvreader)

                # extracting each data row one by one
                for row in csvreader:
                    generate_object_doc(full_doc_dir, fields, row, obj_type)
                    rows.append(row)

                # get total number of rows
                logger.info("Total no. of rows: %d" % (csvreader.line_num))


# if this file is run directly, then start here
if __name__ == '__main__':
    #gen_obj_docs(object_docs)
    configure_overview_table_docs(object_tables)



from enum import Enum
from typing import List

from pydantic import BaseModel

class AttackVersions(Enum):
    V12_1 = "12.1"

class AttackDomains(Enum):
    ENTERPRISE_ATTACK = "enterprise-attack"
    MOBILE_ATTACK = "mobile-attack"
    ICS_ATTACK = "ics-attack"

class ImportType(BaseModel):
    STIX21: bool
    CVE: bool
    identity: bool
    location: bool
    os_threat: bool
    kestrel: bool
    rules: bool
    ATTACK: bool
    ATTACK_Versions: List[AttackVersions]
    ATTACK_Domains: List[AttackDomains]
    CACAO: bool
    US_DoD: bool



class ImportTypeFactory:

    @staticmethod
    def get_import_type_factory():
        return ImportTypeFactory()

    @staticmethod
    def create_import(stix_21=False,
                      attack=False,
                      cve=False,
                      identity=False,
                      location=False,
                      os_threat=False,
                      kestrel=False,
                      rules=False,
                      us_dod=False,
                      attack_versions=[],
                      attack_domains=[],
                      cacao=False):
        if os_threat and (not stix_21 or not attack):
            raise ValueError("os_threat requires stix_21 and attack")

        return ImportType(
            STIX21=stix_21,
            CVE=cve,
            identity=identity,
            location=location,
            os_threat=os_threat,
            kestrel=kestrel,
            rules=rules,
            ATTACK=attack,
            ATTACK_Versions=attack_versions,
            ATTACK_Domains=attack_domains,
            CACAO=cacao,
            US_DoD=us_dod,
        )

    @staticmethod
    def get_attack_import():
        return ImportType(
            STIX21=True,
            CVE=False,
            identity=False,
            location=False,
            os_threat=False,
            kestrel=False,
            rules=False,
            ATTACK=True,
            ATTACK_Versions=[AttackVersions.V12_1],
            ATTACK_Domains=[AttackDomains.ENTERPRISE_ATTACK, AttackDomains.ICS_ATTACK, AttackDomains.MOBILE_ATTACK],
            CACAO=False,
            US_DoD=False
        )

    @staticmethod
    def get_all_imports():
        return ImportType(
            STIX21=True,
            CVE=True,
            identity=True,
            location=True,
            os_threat=True,
            kestrel=True,
            rules=True,
            ATTACK=True,
            ATTACK_Versions=[AttackVersions.V12_1],
            ATTACK_Domains=[AttackDomains.ENTERPRISE_ATTACK, AttackDomains.ICS_ATTACK, AttackDomains.MOBILE_ATTACK],
            CACAO=True,
            US_DoD=True
        )

    @staticmethod
    def get_default_import():
        return ImportType(
            STIX21 = True,
            CVE = False,
            identity = False,
            location = False,
            os_threat = False,
            kestrel = False,
            rules = False,
            ATTACK = False,
            ATTACK_Versions = [AttackVersions.V12_1],
            ATTACK_Domains = [AttackDomains.ENTERPRISE_ATTACK, AttackDomains.ICS_ATTACK, AttackDomains.MOBILE_ATTACK],
            CACAO= False,
            US_DoD= False
        )

    @staticmethod
    def convert_to_dict(
                        import_type: ImportType):
        versions = []
        for version in import_type.ATTACK_Versions:
            versions.append(version.value)

        domains = []
        for domain in import_type.ATTACK_Domains:
            domains.append(domain.value)


        return {
            "STIX21": import_type.STIX21,
            "ATT&CK": import_type.ATTACK,
            "os-threat": import_type.os_threat,
            "kestrel": import_type.kestrel,
            "CACAO": import_type.CACAO,
            "US_DoD": import_type.US_DoD,
            "CVE": import_type.CVE,
            "identity": import_type.identity,
            "location": import_type.location,
            "rules": import_type.rules,
            "ATT&CK_Versions": versions,
            "ATT&CK_Domains": domains
        }





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Loader module for all cacao data model
"""

__author__ = "Brett Forbes"
__credits__ = ["Paolo Di Prodi"]
__license__ = "GPL"
__version__ = "0.1.0"
__maintainer__ = "Paolo Di Prodi"
__email__ = "paolo@priam.ai"
__status__ = "Production"

import json
from glob import glob

import pathlib
from loguru import logger
import os
from pathlib import Path


name = "cacao"
class_model = {}
class_model["sdo"] = {}
class_model["sco"] = {}
class_model["sro"] = {}
class_model["sub"] = {}
class_model["meta"] = {}





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Loader module for all ATT&CK data model
"""

__author__ = "Brett Forbes"
__credits__ = ["Paolo Di Prodi"]
__license__ = "GPL"
__version__ = "0.1.0"
__maintainer__ = "Paolo Di Prodi"
__email__ = "paolo@priam.ai"
__status__ = "Production"




from stixorm.module.definitions.attack.classes import (
    Matrix, Tactic, Technique, SubTechnique, Mitigation, Group, SoftwareMalware,
    SoftwareTool, DataSource, DataComponent, AttackCampaign, Collection,
    ObjectVersion, AttackRelation, AttackMarking, AttackIdentity
)
name = "attack"
class_model={}
class_model["sdo"] = {
    "Matrix": Matrix,
    "Tactic": Tactic,
    "Technique": Technique,
    "SubTechnique": SubTechnique,
    "Mitigation": Mitigation,
    "Group": Group,
    "SoftwareMalware": SoftwareMalware,
    "SoftwareTool": SoftwareTool,
    "DataSource": DataSource,
    "DataComponent": DataComponent,
    "AttackCampaign": AttackCampaign,
    "Collection": Collection,
    "AttackIdentity": AttackIdentity
}
class_model["sub"] = {
    "ObjectVersion": ObjectVersion
}
class_model["sco"] = {}
class_model["meta"] = {
    "AttackMarking": AttackMarking
}
class_model["sro"] = {
    "AttackRelation": AttackRelation
}




import json
import os
import pathlib
from typing import Dict, List
import functools

from stixorm.module.typedb_lib.model.definitions import ObjectKeys

@functools.lru_cache(maxsize=None)
def get_mapping_factory_instance():
    return MappingsFactory()

class MappingsDefinition():

    def __init__(self,
                 name,
                 path):
        self.definition_name = name
        self.definition_path = path

        self.mapping = self.__load_mappings()

    def is_json_file(self, filepath):
        _, file_extension = os.path.splitext(filepath)
        return file_extension.lower() == ".json"

    def __load_definition(self,
                          directory_path: str) -> dict:
        if not os.path.isdir(directory_path):
            return {}

        definitions = {}

        # Iterate over every file in the directory
        for filename in os.listdir(directory_path):
            filepath = os.path.join(directory_path, filename)
            if os.path.isfile(filepath) and self.is_json_file(filepath):
                # Do something with the file, for example, print its contents
                file_name, file_ext = os.path.splitext(filename)
                with open(filepath, 'r') as file:
                    definitions[file_name] = json.load(file)
        return definitions

    def __load_mappings(self) -> dict:

        # Define the directory path
        directory_path = str(self.definition_path.joinpath("mappings"))

        return self.__load_definition(directory_path)

    def get_mappings(self) -> dict:
        return self.mapping

    def get_object_conversion(self) -> list:
        return self.mapping.get("object_conversion", [])
class MappingsFactory():

    def __init__(self):
        self.definition_dir = pathlib.Path(__file__).parents[2].joinpath('definitions')

        self.definitions: Dict[str, MappingsDefinition] = {}

        subfolder_names = self.get_subfolder_names(self.definition_dir)

        for name in subfolder_names:
            self.definitions[name] = MappingsDefinition(name, self.definition_dir.joinpath(name))

    def all_object_keys(self) -> set[ObjectKeys]:
        return set(list(ObjectKeys))

    def all_object_keys_as_string(self) -> set[str]:
        return set([key.value for key in self.all_object_keys()])

    def get_all_types(self) -> List[str]:
        types = []
        for mapping in self.definitions.values():
            for object_conversion in mapping.get_object_conversion():
                if "type" in object_conversion and object_conversion["object"] in self.all_object_keys_as_string():
                    types.append(object_conversion["type"])
        return list(set(types))


    def get_subfolder_names(self, folder_path):
        subfolder_names = []
        for item in os.listdir(folder_path):
            item_path = os.path.join(folder_path, item)
            if os.path.isdir(item_path):
                subfolder_names.append(item)
        return subfolder_names

    def get_mapping_definition(self,
                               name):
        if name in self.definitions.keys():
            return self.definitions.get(name)
        raise Exception(f"Mapping definition {name} not found")


    def get_mappings_from_definition(self,
                                     name):
        return self.get_mapping_definition(name).get_mappings()


import json
from typing import List

import pathlib
from pydantic import BaseModel


class ProcessMap(BaseModel):
    name: str
    keys: List[str]
    match: List[str]
    cond: List[str]

class ProcessMapFactory:

    def __init__(self):
        self.process_map: List[ProcessMap] = []
        self.__load_data()

    @staticmethod
    def process_map_factory() -> 'ProcessMapFactory':
        return ProcessMapFactory()

    # Load process_map.json
    def __load_data(self):
        path = pathlib.Path(__file__).parent.joinpath("process_map.json")
        with open(str(path), 'r') as file:
            data = json.load(file)
        for process_map in data:
            self.process_map.append(ProcessMap(**process_map))

    def all_process_maps(self) -> List[ProcessMap]:
        return self.process_map




import json
import copy
from stixorm.module.authorise import authorised_mappings, import_type_factory
from stix2.exceptions import ParseError
from stix2.parsing import dict_to_stix2
from stixorm.module.parsing.conversion_decisions import sdo_type_to_tql, sro_type_to_tql
import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType

logger = logging.getLogger(__name__)
default_import_type = import_type_factory.get_default_import()


def parse(data: dict, allow_custom=False, import_type: ImportType=default_import_type):
    """Convert a string, dict or file-like object into a STIX object.
    Args:
        data (str, dict, file-like object): The STIX 2 content to be parsed.
        allow_custom (bool): Whether to allow custom properties as well unknown
            custom objects. Note that unknown custom objects cannot be parsed
            into STIX objects, and will be returned as is. Default: False.
        import_type (dict): If present, it describes the set of configuration
            options, which impact the classes the parser will see
    Returns:
        An instantiated Python STIX object.
    Warnings:
        'allow_custom=True' will allow for the return of any supplied STIX
        dict(s) that cannot be found to map to any known STIX object types
        (both STIX2 domain objects or defined custom STIX2 objects); NO
        validation is done. This is done to allow the processing of possibly
        unknown custom STIX objects (example scenario: I need to query a
        third-party TAXII endpoint that could provide custom STIX objects that
        I don't know about ahead of time)
    """
    # convert STIX object to dict, if not already
    logger.debug("i'm in parse, bout to get dict")
    obj = _get_dict(data)
    logger.debug("i'm in parse after get dict")

    # convert dict to full python-stix2 obj
    obj = dict_to_stix(obj, allow_custom, import_type)
    logger.debug(f"############## obj is {obj} ########")

    return obj


def _get_dict(data):
    """Return data as a dictionary.
    Input can be a dictionary, string, or file-like object.
    """

    if type(data) is dict:
        return data
    else:
        try:
            return json.loads(data)
        except TypeError:
            pass
        try:
            return json.load(data)
        except AttributeError:
            pass
        try:
            return dict(data)
        except (ValueError, TypeError):
            raise ValueError(f"Cannot convert {str(data)} to dictionary.")


def is_attack_object(stix_dict):
    if stix_dict.get("x_mitre_domains", False) or stix_dict.get("x_mitre_attack_spec_version", False):
        return True
    else:
        return False


def dict_to_stix(stix_dict: dict,
                 allow_custom=False,
                 import_type: ImportType=default_import_type):
    """convert dictionary to full python-stix2 object
    Args:
        stix_dict (dict): a python dictionary of a STIX object
            that (presumably) is semantically correct to be parsed
            into a full python-stix2 obj
        allow_custom (bool): Whether to allow custom properties as well
            unknown custom objects. Note that unknown custom objects cannot
            be parsed into STIX objects, and will be returned as is.
            Default: False.
        import_type (dict): If present, it describes the set of configuration
            options, which impact the classes the parser will see
    Returns:
        An instantiated Python STIX object
    Warnings:
        'allow_custom=True' will allow for the return of any supplied STIX
        dict(s) that cannot be found to map to any known STIX object types
        (both STIX2 domain objects or defined custom STIX2 objects); NO
        validation is done. This is done to allow the processing of
        possibly unknown custom STIX objects (example scenario: I need to
        query a third-party TAXII endpoint that could provide custom STIX
        objects that I don't know about ahead of time)
    """
    assert len(stix_dict) > 0
    logger.debug(f"I'm in dict to stix, {stix_dict}")
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    if 'type' not in stix_dict:
        raise ParseError(f"Can't parse object with no 'type' property: {str(stix_dict)}")

    obj_type = stix_dict["type"]
    logger.debug(f'\nin parse, raw type is --> {obj_type}')
    logger.debug(f'\n auth-sdo -->{auth["tql_types"]["sdo"]}\n')
    logger.debug(f'\n\n auth-sro -->{auth["tql_types"]["sro"]}\n')
    attack_object = is_attack_object(stix_dict)
    logger.debug(f'attack object {attack_object}')
    #logger.info(f'auth is {auth["tql_types"]["meta"]}')
    if obj_type in auth["types"]["sdo"]:
        logger.debug("Im in sdo")
        sub_technique = False
        if attack_object:
            sub_technique = stix_dict.get("x_mitre_is_subtechnique", False)
        logger.debug(f'subtechnique {sub_technique}, attack {attack_object}')
        obj_tql, sdo_tql_name, is_list, protocol = sdo_type_to_tql(obj_type, import_type, attack_object, sub_technique)
        logger.debug(f"tql name {sdo_tql_name}, obj tql {obj_tql}")
        obj_class = class_for_type(sdo_tql_name, import_type, "sdo")
        logger.debug(f'output  object class is {obj_class}')
    elif obj_type in auth["types"]["sco"]:
        logger.debug("I'm in sco")
        obj_class = class_for_type(obj_type, import_type, "sco")
    elif obj_type in auth["types"]["sro"]:
        logger.debug("I'm in sro")
        uses_relation = False
        is_procedure = False
        if attack_object:
            uses_relation = False if not stix_dict.get("relationship_type", False) == "uses" else True
            is_procedure = False if not stix_dict.get("target_ref", False) == "attack-pattern" else True
        obj_tql = {}
        sro_sub_rel = "" if not stix_dict.get("relationship_type", False) else stix_dict["relationship_type"]

        obj_tql, sro_tql_name, is_list, protocol = sro_type_to_tql(obj_type, sro_sub_rel, import_type, attack_object,
                                                         uses_relation, is_procedure)
        logger.debug(f"~~~~~~~~~~~ sro tql name {sro_tql_name}")
        if obj_type == "relationship":
            if attack_object:
                obj_tql_name = "attack-relation"
            else:
                obj_tql_name = "stix-core-relationship"
        elif obj_type == "sighting":
            obj_tql_name = "sighting"
        if sro_tql_name == "attack-relation":
            obj_tql_name = sro_tql_name
        obj_class = class_for_type(obj_tql_name, import_type, "sro")
    elif obj_type in auth["types"]["sub"]:
        logger.debug("I'm in sub")
        obj_class = class_for_type(obj_type, import_type, "sub")
    elif obj_type in auth["types"]["meta"]:
        logger.debug("I'm in meta")
        if attack_object:
            obj_class = class_for_type("attack-marking", import_type, "meta")
        else:
            obj_class = dict_to_stix2(stix_dict, True)
            logger.debug(f'object class is finally {obj_class}')
            return obj_class

    elif allow_custom:
        logger.debug("I'm in custom")
        raise ParseError(f'the object is not known, and custom is enabled but not implemented')
    else:
        logger.debug("object is not known")
        raise ParseError(f'the object is not known, and custom is not enabled')

    if not obj_class:
        if allow_custom:
            # flag allows for unknown custom objects too, but will not
            # be parsed into STIX object, returned as is
            return stix_dict
        for key_id, ext_def in stix_dict.get('extensions', {}).items():
            if (
                key_id.startswith('extension-definition--') and
                'property-extension' not in ext_def.get('extension_type', '')
            ):
                # prevents ParseError for unregistered objects when
                # allow_custom=False and the extension defines a new object
                return stix_dict
        raise ParseError(f"Can't parse unknown object type {obj_type}! For custom types, use the CustomObject decorator." + str(obj_type))

    logger.debug(f'object class is finally {obj_class}')
    logger.debug("========================================")
    for k, v in stix_dict.items():
        logger.debug(f'k-> {k}, v->{v}')
    logger.debug("=========================================")
    stix_dict['allow_custom'] = allow_custom
    return obj_class(**stix_dict)


def class_for_type(stix_typeql, import_type, category=None):
    """
    Get the registered class which implements a particular STIX type for a
    particular STIX version.
    :param stix_typeql: A STIX type as a string, or for extension-definition
        style extensions, the STIX ID of the definition.
    :param import_type: If present, it describes the set of configuration
            options, which impact the classes the parser will see
    :param category: An optional "category" value, which is just used directly
        as a second key after the STIX version, and depends on how the types
        are internally categorized.  This would be useful if the same STIX type
        is used to mean two different things within the same STIX version.  So
        it's unlikely to be necessary.  Pass None to just search all the
        categories and return the first class found.
    :return: A registered python class which implements the given STIX type, or
        None if one is not found.
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    conv_cls = ""
    cls = None
    logger.debug(f' working out classes, typeql {stix_typeql}, category {category}')

    # find the conversion record
    if category is not None:
        for obj in auth["conv"][category]:
            logger.debug(f'object tql is {obj["typeql"]}, wanted {stix_typeql}')
            if obj["typeql"] == stix_typeql:
                logger.debug("found the right type")
                conv_cls = obj["class"]
                logger.debug(f'classs is {conv_cls}')
                cls = copy.deepcopy(auth["classes"][category][conv_cls])
                logger.debug(f'classs 2 is {cls}')
                return cls

    return cls



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This history all the dictionary definitions and save them as JSON objects
"""

__author__ = "Paolo Di Prodi"
__credits__ = ["Brett Forbes"]
__license__ = "GPL"
__version__ = "0.1.0"
__maintainer__ = "Paolo Di Prodi"
__email__ = "paolo@priam.ai"
__status__ = "Production"

import json
import definitions
import os

os.makedirs('data', exist_ok=True)

for id in dir(definitions):
    obj = getattr(definitions,id)

    # filter the dictionary objects only
    if isinstance(obj, dict) and id!='__builtins__':
        # save the object as dictionary
        with open(f'data/{id}.json', 'w') as file:
            json.dump(obj,file)

    if isinstance(obj, list) and id!='__builtins__':
        # save the object as dictionary
        with open(f'data/{id}.json', 'w') as file:
            json.dump(obj,file)




import logging
import traceback
from enum import Enum
from typing import Optional, List, Dict

from networkx import DiGraph, find_cycle, contracted_nodes, topological_sort
from pydantic import BaseModel


from stixorm.module.typedb_lib.logging import log_insert_query

class TypeQLObject(BaseModel):
    dep_match: str
    dep_insert: str
    indep_ql: str
    core_ql: str
    dep_list: List[str]


class ResultStatus(Enum):
    SUCCESS = "success"
    ERROR = "error"
    ALREADY_IN_DB = "already_in_db"
    UNKNOWN = "unknown"
    CYCLICAL_DEPENDENCY = "cyclical_dependency"
    MISSING_DEPENDENCY = "missing_Dependency"
    VALID_FOR_DB_COMMIT = "valid_for_db_commit"

class Result(BaseModel):
    id: str
    status: ResultStatus
    error: Optional[str]
    message: Optional[str]


class Instructions:

    def __init__(self):
        self.instructions: Dict[str, Instruction] = {}
        self.dependencies: Optional[DiGraph] = None
        self.verified_missing_dependencies = []
        self.order = []
        self.cyclical_references = {}


    def convert_to_result(self):
        results = []
        message = None
        error = None
        status = ResultStatus.UNKNOWN
        instruction: Instruction
        for instruction in self.instructions.values():
            if instruction.status in [Status.SUCCESS]:
                status = ResultStatus.SUCCESS
            elif instruction.status == Status.EXCLUDE_EXISTS_IN_DATABASE:
                status = ResultStatus.ALREADY_IN_DB
            elif instruction.status in [Status.ERROR]:
                status = ResultStatus.ERROR
                error = instruction.error
            elif instruction.status == Status.FAILED_CYCLICAL:
                status = ResultStatus.CYCLICAL_DEPENDENCY
            elif instruction.status == Status.FAILED_MISSING_DEPENDENCY:
                status = ResultStatus.MISSING_DEPENDENCY
                message = "Missing values " + str(instruction.missing)
            elif instruction.status in [Status.CREATED_QUERY, Status.CREATED]:
                status = ResultStatus.VALID_FOR_DB_COMMIT


            results.append(Result(id=instruction.id, error=error, status=status, message=message))
        return results

    def add_dependencies(self, dependencies: DiGraph):
        self.dependencies = dependencies

    def not_allow_insertion(self,
                            id: str):
        return self.instructions[id].status != Status.CREATED_QUERY

    def __create_key(self, string1, string2):
        concatenated_string = [string1 , string2]
        sorted_string = ''.join(sorted(concatenated_string))
        return sorted_string

    def compress_cyclical_ids(self):
        for tuple in self.cyclical_ids():
            id = self.__create_key(tuple[0], tuple[1])
            if id in self.cyclical_references:
                continue
            new_graph = merge(self.dependencies, tuple[0], tuple[1])
            self.dependencies = new_graph
            self.cyclical_references[id] = [tuple[0], tuple[1]]

    def cyclical_ids(self):
        try:
            return find_cycle(self.dependencies, orientation="original")
        except Exception as e:
            return []

    def update_ids_in_database(self,
                               ids: List[str]):
        for instruction in self.instructions.values():
            if instruction.id in ids:
                instruction.status = Status.EXCLUDE_EXISTS_IN_DATABASE

    def exist_cyclical_ids(self):
        return len(self.cyclical_ids()) > 0

    def add_insertion_order(self,
                            order):
        included_ids_in_order = self.instructions.keys()
        # only order for
        for id in order:
            if id in included_ids_in_order:
                self.order.append(id)


    def missing_dependency_ids(self):
        missing = []

        ids_in_instructions = []
        for instruction in self.instructions.values():
            ids_in_instructions.append(instruction.id)

        for node in self.dependencies.nodes:
            if node not in ids_in_instructions:
                missing.append(node)

        return missing

    def exist_missing_dependencies(self):
        return len(self.verified_missing_dependencies) > 0

    def register_missing_dependencies(self,
                                      missing: List[str]):
        self.verified_missing_dependencies = missing
        instruction: AddInstruction
        for instruction in self.instructions.values():
            if instruction.status == Status.ERROR:
                continue
            missing_dependencies = set(instruction.typeql_obj.dep_list).intersection(set(missing))
            if len(missing_dependencies) > 0:
                instruction.status = Status.FAILED_MISSING_DEPENDENCY
                instruction.missing = list(missing_dependencies)

    def register_cyclical_dependencies(self):
        instruction: AddInstruction
        for instruction in self.instructions.values():
            if instruction.status == Status.ERROR:
                continue
            cyclical_ids = [item for tup in self.cyclical_ids() for item in tup]
            if instruction.id in set(cyclical_ids):
                instruction.status = Status.FAILED_CYCLICAL

    def create_insert_queries(self,
                              build_insert_query):
        instruction: Instruction
        for instruction in self.instructions.values():
            if instruction.status != Status.CREATED:
                continue
            try:
                result = build_insert_query(instruction.typeql_obj.dict())
                instruction.status = Status.CREATED_QUERY
                instruction.query = result
            except Exception as e:
                instruction.status = Status.ERROR
                instruction.error = str(e)
                log_insert_query(result, instruction.layer)



    def get_ordered_ids(self):
        if len(self.order) == 0:
            return self.instructions.keys()

        return self.order

    def update_instruction_as_success(self,
                                      id: str):
        self.instructions[id].status = Status.SUCCESS

    def update_delete_instruction_as_success(self,
                                      id: str):
        self.instructions[id].status = Status.SUCCESS

    def update_instruction_as_error(self,
                                    id: str,
                                    error: str):
        self.instructions[id].status = Status.ERROR
        self.instructions[id].error = error

    def update_delete_instruction_as_error(self,
                                    id: str,
                                    error: str):
        self.instructions[id].status = Status.ERROR
        self.instructions[id].error = error

    def get_query_for_id(self,
                         id: str):
        return self.instructions[id].query

    def insert_add_instruction(self,
                               id: str,
                               typeql_obj: Optional[TypeQLObject]):
        self.instructions[id] = AddInstruction(status=Status.CREATED, id=id, typeql_obj=typeql_obj)

    def insert_delete_instruction(self,
                               id: str,
                               layer: dict):
        self.instructions[id] = DeleteInstruction(status=Status.CREATED_QUERY,
                                                id=id,
                                                layer=layer,
                                                query=layer['delete'])

    def insert_delete_instruction_error(self,
                                 id: str,
                                 error: Exception):
        if error is None:
            return
        logging.exception("\n".join(traceback.format_exception(error)))
        self.instructions[id] = DeleteInstruction(status=Status.ERROR, id=id, error=str(error))

    def insert_instruction_error(self,
                                 id: str,
                                 error: Exception):
        if error is None:
            return
        logging.exception("\n".join(traceback.format_exception(error)))
        self.instructions[id] = AddInstruction(status=Status.ERROR, id=id, error=str(error))

    def insert_add_insert_missing_dependency(self,
                                             id: str,
                                             typeql_obj: TypeQLObject):
        self.instructions[id] = AddInstruction(status=Status.FAILED_MISSING_DEPENDENCY, id=id, typeql_obj= typeql_obj)

    def insert_add_instruction_cyclical(self,
                                        id: str,
                                        layer: dict):
        self.instructions[id] = AddInstruction(status=Status.FAILED_CYCLICAL, id=id, layer=layer)


class Status(Enum):
    ERROR = 'error'
    SUCCESS = "success"
    CREATED_QUERY = 'created_query'
    EXCLUDE_EXISTS_IN_DATABASE = 'exists_in_database'
    FAILED_MISSING_DEPENDENCY = 'missing_dependency'
    FAILED_CYCLICAL = 'cyclical'
    CREATED= "created"










class AddInfo(BaseModel):
    status: Status
    id: str
    error: Optional[str]  = None

class Instruction(BaseModel):
    status: Status
    id: str
    layer: Optional[dict]  = None
    query: Optional[str]  = None
    error: Optional[str]  = None
    missing: Optional[List[str]] = None

class AddInstruction(Instruction):
    typeql_obj: Optional[TypeQLObject]

class DeleteInstruction(Instruction):
    pass

class AddInstructions(BaseModel):
    order_instructions: List[AddInstruction]
    cyclical: List[AddInstruction]
    errors: List[AddInstruction]
    missing: List[str]
